/* ICMP Redirect control by sysctl function.
 * Copyright (C) 2005 Yasuhiro Ohara
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <zebra.h>

#include "log.h"
#include "vty.h"
#include "command.h"

#define MIB_SIZ 4
#define VNL     VTY_NEWLINE

int
sysctl_get (int *mib)
{
  int len;
  int value = 0;

  len = sizeof (value);
  if (sysctl (mib, MIB_SIZ, &value, &len, 0, 0) < 0) 
    {
      zlog_warn ("Can't get sysctl value");
      return -1;
    }
  return value;
}

int
sysctl_on (int *mib)
{
  int len;
  int on = 1;

  len = sizeof (on);
  if (sysctl (mib, MIB_SIZ, NULL, NULL, &on, len) < 0) 
    {
      zlog_warn ("Can't set sysctl on");
      return -1;
    }
  return on;
}

int
sysctl_off (int *mib)
{
  int len;
  int off = 0;

  len = sizeof (off);
  if (sysctl (mib, MIB_SIZ, NULL, NULL, &off, len) < 0) 
    {
      zlog_warn ("Can't set sysctl off");
      return -1;
    }
  return off;
}

int ipredirect[MIB_SIZ] =
{
  CTL_NET,
  PF_INET,
  IPPROTO_IP,
  IPCTL_SENDREDIRECTS
};

#ifdef ICMPCTL_REDIRACCEPT
int icmprediraccept[MIB_SIZ] =
{
  CTL_NET,
  PF_INET,
  IPPROTO_ICMP,
  ICMPCTL_REDIRACCEPT
};
#endif /*ICMPCTL_REDIRACCEPT*/

DEFUN (show_ip_redirect,
       show_ip_redirect_cmd,
       "show ip redirect (send|accept)",
       SHOW_STR
       IP_STR
       "ICMP redirecting status\n")
{
  int ret = 1;
  if (! strncmp (argv[0], "se", 2))
    ret = sysctl_get (ipredirect);
#ifdef ICMPCTL_REDIRACCEPT
  else if (! strncmp (argv[0], "ac", 2))
    ret = sysctl_get (icmprediraccept);
#endif /*ICMPCTL_REDIRACCEPT*/

  if (ret == 1)
    vty_out (vty, "ICMP redirect %s is on%s", argv[0], VNL);
  else
    vty_out (vty, "ICMP redirect %s is off%s", argv[0], VNL);

  return CMD_SUCCESS;
}

DEFUN (ip_redirect,
       ip_redirect_cmd,
       "ip redirect (send|accept)",
       IP_STR
       "ICMP Redirect Control\n"
       "Send ICMP Redirect\n"
       "Accept ICMP Redirect\n")
{
  int ret = 0;
  if (! strncmp (argv[0], "se", 2))
    ret = sysctl_on (ipredirect);
#ifdef ICMPCTL_REDIRACCEPT
  else if (! strncmp (argv[0], "ac", 2))
    ret = sysctl_on (icmprediraccept);
#endif /*ICMPCTL_REDIRACCEPT*/

  if (ret != 1)
    {
      vty_out (vty, "Can't turn on ICMP redirect %s%s", argv[0], VNL);
      return CMD_WARNING;
    }

  return CMD_SUCCESS;
}

DEFUN (no_ip_redirect,
       no_ip_redirect_cmd,
       "no ip redirect (send|accept)",
       NO_STR
       IP_STR
       "ICMP Redirect Control\n"
       "Send ICMP Redirect\n"
       "Accept ICMP Redirect\n")
{
  int ret = 0;
  if (! strncmp (argv[0], "se", 2))
    ret = sysctl_off (ipredirect);
#ifdef ICMPCTL_REDIRACCEPT
  else if (! strncmp (argv[0], "ac", 2))
    ret = sysctl_off (icmprediraccept);
#endif /*ICMPCTL_REDIRACCEPT*/

  if (ret != 0)
    {
      vty_out (vty, "Can't turn off ICMP redirect %s%s", argv[0], VNL);
      return CMD_WARNING;
    }

  return CMD_SUCCESS;
}


#ifdef HAVE_IPV6
int ip6redirect[MIB_SIZ] =
{
  CTL_NET,
  PF_INET6,
  IPPROTO_IPV6,
  IPV6CTL_SENDREDIRECTS
};

int icmp6rediraccept[MIB_SIZ] =
{
  CTL_NET,
  PF_INET6,
  IPPROTO_ICMPV6,
  ICMPV6CTL_REDIRACCEPT
};

DEFUN (show_ipv6_redirect,
       show_ipv6_redirect_cmd,
       "show ipv6 redirect (send|accept)",
       SHOW_STR
       IP_STR
       "ICMP redirecting status\n")
{
  int ret = 0;
  if (! strncmp (argv[0], "se", 2))
    ret = sysctl_get (ip6redirect);
  else if (! strncmp (argv[0], "ac", 2))
    ret = sysctl_get (icmp6rediraccept);

  if (ret == 1)
    vty_out (vty, "ICMPv6 redirect %s is on%s", argv[0], VNL);
  else
    vty_out (vty, "ICMPv6 redirect %s is off%s", argv[0], VNL);

  return CMD_SUCCESS;
}

DEFUN (ipv6_redirect,
       ipv6_redirect_cmd,
       "ipv6 redirect (send|accept)",
       IPV6_STR
       "ICMPv6 Redirect Control\n"
       "Send ICMPv6 Redirect\n"
       "Accept ICMPv6 Redirect\n")
{
  int ret = 0;
  if (! strncmp (argv[0], "se", 2))
    ret = sysctl_on (ip6redirect);
  else if (! strncmp (argv[0], "ac", 2))
    ret = sysctl_on (icmp6rediraccept);

  if (ret != 1)
    {
      vty_out (vty, "Can't turn on ICMPv6 redirect %s%s", argv[0], VNL);
      return CMD_WARNING;
    }

  return CMD_SUCCESS;
}

DEFUN (no_ipv6_redirect,
       no_ipv6_redirect_cmd,
       "no ipv6 redirect (send|accept)",
       NO_STR
       IPV6_STR
       "ICMPv6 Redirect Control\n"
       "Send ICMPv6 Redirect\n"
       "Accept ICMPv6 Redirect\n")
{
  int ret = 0;
  if (! strncmp (argv[0], "se", 2))
    ret = sysctl_off (ip6redirect);
  else if (! strncmp (argv[0], "ac", 2))
    ret = sysctl_off (icmp6rediraccept);

  if (ret != 0)
    {
      vty_out (vty, "Can't turn off ICMPv6 redirect %s%s", argv[0], VNL);
      return CMD_WARNING;
    }

  return CMD_SUCCESS;
}

#endif /* HAVE_IPV6 */

int
config_write_redirect (struct vty *vty)
{
  if (! sysctl_get (ipredirect))
    vty_out (vty, "no ip redirect send%s", VNL);
#ifdef ICMPCTL_REDIRACCEPT
  if (! sysctl_get (icmprediraccept))
    vty_out (vty, "no ip redirect accept%s", VNL);
#endif /*ICMPCTL_REDIRACCEPT*/
#ifdef HAVE_IPV6
  if (! sysctl_get (ip6redirect))
    vty_out (vty, "no ipv6 redirect send%s", VNL);
  if (! sysctl_get (icmp6rediraccept))
    vty_out (vty, "no ipv6 redirect accept%s", VNL);
#endif /* HAVE_IPV6 */
  vty_out (vty, "!%s", VTY_NEWLINE);
  return 0;
}

/* table node for routing tables. */
struct cmd_node redirect_node =
{
  REDIRECT_NODE,
  "",				/* This node has no interface. */
  1
};

void
redirect_init ()
{
  install_node (&redirect_node, config_write_redirect);

  install_element (VIEW_NODE, &show_ip_redirect_cmd);
  install_element (ENABLE_NODE, &show_ip_redirect_cmd);
  install_element (CONFIG_NODE, &ip_redirect_cmd);
  install_element (CONFIG_NODE, &no_ip_redirect_cmd);

#ifdef HAVE_IPV6
  install_element (VIEW_NODE, &show_ipv6_redirect_cmd);
  install_element (ENABLE_NODE, &show_ipv6_redirect_cmd);
  install_element (CONFIG_NODE, &ipv6_redirect_cmd);
  install_element (CONFIG_NODE, &no_ipv6_redirect_cmd);
#endif /* HAVE_IPV6 */
}


